@@ -1,4 +1,4 @@
-from typing import Callable
+from typing import Awaitable, Callable
 import asyncio
 import logging
 
@@ -22,33 +22,18 @@ class async_runner:
         self.on_closes.append(fn)
         return fn
 
-    async def close_check_task(self):
-        while True:
-            try:
-                has_no_finished_task = False
-                cur_task = asyncio.current_task(self.loop)
-                for task in asyncio.all_tasks(self.loop):
-                    if cur_task == task or task.done():
-                        continue
-                    has_no_finished_task = True
-                    break
-
-                if not has_no_finished_task:
-                    return
-
-                await asyncio.sleep(1, loop=self.loop)
-            except:
-                pass
-
-    def run_forever(self, loop: asyncio.BaseEventLoop = asyncio.get_event_loop()):
-        self.loop = loop
+    def run(self, loop: asyncio.BaseEventLoop = asyncio.get_event_loop(), main_task: Awaitable = None):
         try:
-            self.loop.run_forever()
+            if main_task:
+                loop.run_until_complete(main_task)
+            else:
+                loop.run_forever()
         except KeyboardInterrupt:
             self.exeute_on_close()
-            self.loop.run_until_complete(self.close_check_task())
-            self.loop.stop()
-            self.loop.close()
+            for task in asyncio.all_tasks(loop):
+                task.cancel()
+            loop.run_until_complete(loop.shutdown_asyncgens())
+            loop.close()
         except BaseException as e:
             logging.exception(e.with_traceback(None), stack_info=True)
             return
@@ -59,3 +44,7 @@ g_async_runner = async_runner()
 
 def on_close_function(fn: Callable):
     return g_async_runner.on_close_function(fn)
+
+
+def run(loop: asyncio.BaseEventLoop = asyncio.get_event_loop(), main_task: Awaitable = None):
+    g_async_runner.run(loop, main_task)
